from typing import List, Tuple, Union

import graphql
from hypothesis import strategies as st
from hypothesis.errors import InvalidArgument

from ..cache import cached_build_schema
from ..types import CustomScalarStrategies


def maybe_parse_schema(schema: Union[str, graphql.GraphQLSchema]) -> graphql.GraphQLSchema:
    if isinstance(schema, str):
        return cached_build_schema(schema)
    return schema


def validate_fields(fields_: Tuple[str, ...], available_fields: List[str]) -> None:
    if not fields_:
        raise ValueError("If you pass `fields`, it should not be empty")
    invalid_fields = tuple(field for field in fields_ if field not in available_fields)
    if invalid_fields:
        raise ValueError(f"Unknown fields: {', '.join(invalid_fields)}")


def validate_custom_scalars(strategies: CustomScalarStrategies) -> None:
    assert isinstance(strategies, dict)
    for name, strategy in strategies.items():
        validate_scalar_strategy(name, strategy)


def validate_scalar_strategy(name: str, strategy: st.SearchStrategy) -> None:
    if not isinstance(name, str):
        raise InvalidArgument(f"scalar name {name!r} must be a string")
    if not isinstance(strategy, st.SearchStrategy):
        raise InvalidArgument(
            f"custom_scalars[{name!r}]={strategy!r} must be a Hypothesis "
            "strategy which generates AST nodes matching this scalar."
        )



import graphql

from ..types import SelectionNodes


def make_document_node(selections: SelectionNodes, *, kind: graphql.OperationType) -> graphql.DocumentNode:
    """Create top-level node for an operation AST."""
    return graphql.DocumentNode(
        kind="document",
        definitions=[
            graphql.OperationDefinitionNode(
                kind="operation_definition",
                operation=kind,
                selection_set=graphql.SelectionSetNode(kind="selection_set", selections=selections),
            )
        ],
    )


def make_query(selections: SelectionNodes) -> graphql.DocumentNode:
    return make_document_node(selections, kind=graphql.OperationType.QUERY)


def make_mutation(selections: SelectionNodes) -> graphql.DocumentNode:
    return make_document_node(selections, kind=graphql.OperationType.MUTATION)



import concurrent.futures
import json
import urllib.request
from concurrent.futures._base import Future
from contextlib import suppress
from typing import Any, Dict, List

import graphql

API_DIRECTORY_URL = "https://raw.githubusercontent.com/APIs-guru/graphql-apis/master/apis.json"
FETCH_TIMEOUT = 30
CORPUS_PATH = "test/corpus-api-guru-catalog.json"
INTROSPECTION_QUERY = graphql.get_introspection_query()
INTROSPECTION_QUERY_PAYLOAD = json.dumps({"query": INTROSPECTION_QUERY}).encode("utf8")
INVALID_API_URLS = {"Deutsche Bahn": "https://bahnql.herokuapp.com/graphql"}
ADDITIONAL_APIS = {
    "AniList": "https://graphql.anilist.co/",
    "Artsy": "https://metaphysics-production.artsy.net/",
    "Contentful": "https://graphql.contentful.com/content/v1/spaces/f8bqpb154z8p/environments/master"
    "?access_token=9d5de88248563ebc0d2ad688d0473f56fcd31c600e419d6c8962f6aed0150599",
    "Countries v2": "https://countries-274616.ew.r.appspot.com/",
    "EtMDB": "https://etmdb.com/graphql",
    "Gitlab": "https://gitlab.com/api/graphql",
    "GraphQL Jobs": "https://api.graphql.jobs/",
    "React Finland": "https://api.react-finland.fi/graphql",
    "Universe": "https://www.universe.com/graphql",
    "TravelgateX": "https://api.travelgatex.com/",
    "Barcelona Urban Mobility API": "https://barcelona-urban-mobility-graphql-api.netlify.app/graphql",
    "TMDB": "https://tmdb.apps.quintero.io/",
    "SWAPI": "https://swapi-graphql.netlify.app/.netlify/functions/index",
    "Spotify": "https://spotify-api-graphql-console.herokuapp.com/graphql",
    "Spacex Land": "https://api.spacex.land/graphql",
    "PokeAPI": "https://pokeapi-graphiql.herokuapp.com/",
    "MusicBrainz": "https://graphbrainz.herokuapp.com/",
    "Ghibliql": "https://ghibliql.herokuapp.com/",
    "Câmara dos deputados Brasil": "https://graphql-camara-deputados.herokuapp.com/",
    "Weather API": "https://graphql-weather-api.herokuapp.com/",
    "UN SDG data series API": "http://linkedsdg.apps.officialstatistics.org/graphql/linkedsdg.apps.officialstatistics.org/graphql",
    "The Rick and Morty API": "https://rickandmortyapi.com/graphql",
    "Google directions API": "https://directions-graphql.herokuapp.com/graphql",
    "Spotify GraphQL Server": "https://spotify-graphql-server.herokuapp.com/graphql",
    "Planets": "https://pristine-gadget-267405.appspot.com/graphql/",
    "MongoDB Northwind demo": "https://graphql-compose.herokuapp.com/northwind/",
}


def load_json(url: str) -> Any:
    with urllib.request.urlopen(url) as response:
        return json.load(response)


def load_schema(url: str) -> str:
    request = urllib.request.Request(
        url,
        data=INTROSPECTION_QUERY_PAYLOAD,
        headers={"Content-Type": "application/json"},
    )
    with urllib.request.urlopen(request) as response:
        data = json.load(response)
    return introspection_result_to_sdl(data["data"])


def introspection_result_to_sdl(data: Dict[str, Any]) -> str:
    client_schema = graphql.build_client_schema(data)
    return graphql.print_schema(client_schema).strip()


def load_corpus(path: str = CORPUS_PATH) -> Dict[str, Any]:
    try:
        with open(path) as fd:
            return json.load(fd)
    except FileNotFoundError:
        return {}


def store_corpus(data: Dict[str, Any], path: str = CORPUS_PATH) -> None:
    with open(path, mode="w") as fd:
        json.dump(data, fd, indent=4, sort_keys=True)


def update_corpus(futures: List[Future], corpus_path: str = CORPUS_PATH) -> Dict[str, Any]:
    corpus = load_corpus(corpus_path)
    with suppress(concurrent.futures.TimeoutError):
        for future in concurrent.futures.as_completed(futures, timeout=FETCH_TIMEOUT):
            try:
                schema = future.result()
                corpus[future.name] = schema
            except Exception as exc:
                print(f"Error: {future.name} {exc}")
    return corpus


def main(directory_url: str = API_DIRECTORY_URL, corpus_path: str = CORPUS_PATH):
    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = []

        def submit_future(_name, _url):
            future = executor.submit(load_schema, _url)
            future.name = _name
            futures.append(future)

        for api in load_json(directory_url):
            name = api["info"]["title"]
            if name in INVALID_API_URLS:
                url = INVALID_API_URLS[name]
            else:
                url = api["url"]

            submit_future(name, url)
        for name, url in ADDITIONAL_APIS.items():
            submit_future(name, url)

        corpus = update_corpus(futures, corpus_path)

    store_corpus(corpus, corpus_path)


if __name__ == "__main__":
    main()



import pytest
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.errors import InvalidArgument

from hypothesis_graphql import mutations


@pytest.fixture(scope="session")
def mutation(schema):
    return f"""{schema}
type Query {{
  getBooks: [Book]
  getAuthors: [Author]
}}
type Mutation {{
    addBook(title: String!, author: String!): Book!
    addAuthor(name: String!): Author!
    }}"""


ALL_FIELDS = {"addBook", "addAuthor"}


@pytest.mark.parametrize("fields", (None, {"addBook"}, {"addAuthor"}))
@given(data=st.data())
def test_mutation(data, mutation, validate_operation, fields):
    mut = data.draw(mutations(mutation, fields=fields))
    if fields:
        not_selected_fields = ALL_FIELDS - fields
        for field in not_selected_fields:
            assert field not in mut
    validate_operation(mutation, mut)


def test_no_mutation(schema):
    with pytest.raises(InvalidArgument, match="Mutation type is not defined in the schema"):
        mutations(schema)



import json
import pathlib
from dataclasses import dataclass
from typing import Optional

import graphql
import pytest
from hypothesis import HealthCheck, Phase, Verbosity, given, settings
from hypothesis import strategies as st

from hypothesis_graphql import from_schema, nodes
from hypothesis_graphql._strategies.strategy import BUILT_IN_SCALAR_TYPE_NAMES
from hypothesis_graphql.cache import cached_build_schema

HERE = pathlib.Path(__file__).parent

with open(HERE / "corpus-api-guru-catalog.json") as fd:
    schemas = json.load(fd)

INVALID_SCHEMAS = {
    # Error: The directive '@deprecated' can only be used once at this location
    "Gitlab",
}


@dataclass
class Schema:
    raw: dict
    custom_scalars: Optional[dict]


PLACEHOLDER_STRATEGY = st.just("placeholder").map(nodes.String)


@pytest.fixture
def schema(request):
    raw_schema = schemas[request.param]
    parsed = cached_build_schema(raw_schema)
    custom_scalars = {}
    # Put placeholders for every custom scalar. Their value is pretty much irrelevant here, it is more important to
    # test query generation, therefore a placeholder will allow these tests to run on schemas that contain custom
    # scalars
    for name, type_ in parsed.type_map.items():
        if name not in BUILT_IN_SCALAR_TYPE_NAMES and isinstance(type_, graphql.GraphQLScalarType):
            custom_scalars[name] = PLACEHOLDER_STRATEGY
    return Schema(raw_schema, custom_scalars or None)


def get_names(corpus, predicate=None):
    for name in sorted(corpus):
        if name in INVALID_SCHEMAS or (predicate and not predicate(name)):
            continue
        yield name


CORPUS_SETTINGS = {
    "suppress_health_check": [
        HealthCheck.too_slow,
        HealthCheck.data_too_large,
        HealthCheck.filter_too_much,
        HealthCheck.function_scoped_fixture,
    ],
    "phases": [Phase.generate],
    "verbosity": Verbosity.quiet,
    "deadline": None,
    "max_examples": 10,
}


@pytest.mark.parametrize("schema", get_names(schemas), indirect=["schema"])
@settings(**CORPUS_SETTINGS)
@given(data=st.data())
def test_corpus(data, schema: Schema, validate_operation):
    query = data.draw(from_schema(schema.raw, custom_scalars=schema.custom_scalars))
    validate_operation(schema.raw, query)



"""Strategies for simple types like scalars or enums."""

from functools import lru_cache
from typing import Optional, Tuple, Type, TypeVar, Union

import graphql
from hypothesis import strategies as st
from hypothesis.errors import InvalidArgument

from .. import nodes
from ..types import ScalarValueNode

T = TypeVar("T")
MIN_INT = -(2**31)
MAX_INT = 2**31 - 1


# `String` version without extra `str` call
def _string(
    value: str, StringValueNode: Type[graphql.StringValueNode] = graphql.StringValueNode
) -> graphql.StringValueNode:
    return StringValueNode(value=value)


INTEGER_STRATEGY = st.integers(min_value=MIN_INT, max_value=MAX_INT).map(nodes.Int)
FLOAT_STRATEGY = st.floats(allow_infinity=False, allow_nan=False).map(nodes.Float)
BOOLEAN_STRATEGY = st.booleans().map(nodes.Boolean)
NULL_STRATEGY = st.just(nodes.Null)


@lru_cache(maxsize=16)
def scalar(
    alphabet: st.SearchStrategy[str],
    type_name: str,
    nullable: bool = True,
    default: Optional[graphql.ValueNode] = None,
) -> st.SearchStrategy[ScalarValueNode]:
    if type_name == "Int":
        return int_(nullable=nullable, default=default)
    if type_name == "Float":
        return float_(nullable=nullable, default=default)
    if type_name == "String":
        return string(nullable=nullable, default=default, alphabet=alphabet)
    if type_name == "ID":
        return id_(nullable=nullable, default=default, alphabet=alphabet)
    if type_name == "Boolean":
        return boolean(nullable=nullable, default=default)
    raise InvalidArgument(
        f"Scalar {type_name!r} is not supported. "
        "Provide a Hypothesis strategy via the `custom_scalars` argument to generate it."
    )


def int_(
    *, nullable: bool = True, default: Optional[graphql.ValueNode] = None
) -> st.SearchStrategy[graphql.IntValueNode]:
    return maybe_default(maybe_null(INTEGER_STRATEGY, nullable), default=default)


def float_(
    *, nullable: bool = True, default: Optional[graphql.ValueNode] = None
) -> st.SearchStrategy[graphql.FloatValueNode]:
    return maybe_default(maybe_null(FLOAT_STRATEGY, nullable), default=default)


def string(
    *, nullable: bool = True, default: Optional[graphql.ValueNode] = None, alphabet: st.SearchStrategy[str]
) -> st.SearchStrategy[graphql.StringValueNode]:
    return maybe_default(
        maybe_null(st.text(alphabet=alphabet).map(_string), nullable),
        default=default,
    )


def id_(
    *, nullable: bool = True, default: Optional[graphql.ValueNode] = None, alphabet: st.SearchStrategy[str]
) -> st.SearchStrategy[Union[graphql.StringValueNode, graphql.IntValueNode]]:
    return maybe_default(string(nullable=nullable, alphabet=alphabet) | int_(nullable=nullable), default=default)


def boolean(
    *, nullable: bool = True, default: Optional[graphql.ValueNode] = None
) -> st.SearchStrategy[graphql.BooleanValueNode]:
    return maybe_default(maybe_null(BOOLEAN_STRATEGY, nullable), default=default)


def maybe_null(strategy: st.SearchStrategy[T], nullable: bool) -> st.SearchStrategy[T]:
    if nullable:
        strategy |= NULL_STRATEGY
    return strategy


def custom(
    strategy: st.SearchStrategy,
    nullable: bool = True,
    default: Optional[graphql.ValueNode] = None,
) -> st.SearchStrategy:
    return maybe_default(maybe_null(strategy, nullable), default=default)


def maybe_default(
    strategy: st.SearchStrategy[T], *, default: Optional[graphql.ValueNode] = None
) -> st.SearchStrategy[T]:
    if default is not None:
        strategy |= st.just(default)
    return strategy


@lru_cache(maxsize=64)
def enum(
    values: Tuple[str],
    nullable: bool = True,
    default: Optional[graphql.ValueNode] = None,
) -> st.SearchStrategy[graphql.EnumValueNode]:
    return maybe_default(maybe_null(st.sampled_from(values).map(nodes.Enum), nullable), default=default)


def list_(
    strategy: st.SearchStrategy[graphql.ListValueNode],
    nullable: bool = True,
    default: Optional[graphql.ValueNode] = None,
) -> st.SearchStrategy[graphql.ListValueNode]:
    return maybe_default(maybe_null(strategy.map(nodes.List), nullable), default=default)



from typing import Callable, Dict, List, Union

import graphql
from hypothesis import strategies as st

# Leaf nodes, that don't have children
ScalarValueNode = Union[
    graphql.IntValueNode,
    graphql.FloatValueNode,
    graphql.StringValueNode,
    graphql.BooleanValueNode,
]
InputTypeNode = Union[
    ScalarValueNode,
    graphql.EnumValueNode,
    graphql.ListValueNode,
    graphql.ObjectValueNode,
]
Field = Union[graphql.GraphQLField, graphql.GraphQLInputField]
InterfaceOrObject = Union[graphql.GraphQLObjectType, graphql.GraphQLInterfaceType]
SelectionNodes = List[graphql.SelectionNode]
AstPrinter = Callable[[graphql.Node], str]
CustomScalarStrategies = Dict[str, st.SearchStrategy[graphql.ValueNode]]



# pylint: disable=unused-import
import dataclasses
import operator
from functools import reduce, wraps
from operator import or_
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import graphql
from graphql import is_equal_type
from hypothesis import strategies as st
from hypothesis.errors import InvalidArgument
from hypothesis.strategies._internal.utils import cacheable

from .. import nodes
from ..types import (
    AstPrinter,
    CustomScalarStrategies,
    Field,
    InputTypeNode,
    InterfaceOrObject,
    SelectionNodes,
)
from . import factories, primitives, validation
from .ast import make_mutation, make_query
from .containers import flatten

BY_NAME = operator.attrgetter("name")
EMPTY_LISTS_STRATEGY = st.builds(list)
BUILT_IN_SCALAR_TYPE_NAMES = {"Int", "Float", "String", "ID", "Boolean"}


def instance_cache(key_func: Callable) -> Callable:
    def decorator(method: Callable) -> Callable:
        @wraps(method)
        def wrapped(self: "GraphQLStrategy", *args: Any, **kwargs: Any) -> st.SearchStrategy:
            key = key_func(*args, **kwargs)
            memo = self._cache.setdefault(method.__name__, {})
            cached = memo.get(key)
            if cached is not None:
                return cached
            result = method(self, *args, **kwargs)
            memo[key] = result
            return result

        return wrapped

    return decorator


@dataclasses.dataclass
class GraphQLStrategy:
    """Strategy for generating various GraphQL nodes."""

    schema: graphql.GraphQLSchema
    alphabet: st.SearchStrategy[str]
    custom_scalars: CustomScalarStrategies = dataclasses.field(default_factory=dict)
    allow_null: bool = True
    # As the schema is assumed to be immutable, there are a few strategy caches possible for internal components
    # This is a per-method cache without limits as they are proportionate to the schema size
    _cache: Dict[str, Dict] = dataclasses.field(default_factory=dict)

    def values(
        self,
        type_: graphql.GraphQLInputType,
        default: Optional[graphql.ValueNode] = None,
    ) -> st.SearchStrategy[InputTypeNode]:
        """Generate value nodes of a type, that corresponds to the input type.

        They correspond to all `GraphQLInputType` variants:
            - GraphQLScalarType -> ScalarValueNode
            - GraphQLEnumType -> EnumValueNode
            - GraphQLInputObjectType -> ObjectValueNode

        GraphQLWrappingType[T] is unwrapped:
            - GraphQLList -> ListValueNode[T]
            - GraphQLNonNull -> T (processed with nullable=False)
        """
        type_, nullable = check_nullable(type_)
        if not self.allow_null:
            nullable = False
        # Types without children
        if isinstance(type_, graphql.GraphQLScalarType):
            type_name = type_.name
            if type_name in self.custom_scalars:
                return primitives.custom(self.custom_scalars[type_name], nullable, default=default)
            return primitives.scalar(alphabet=self.alphabet, type_name=type_name, nullable=nullable, default=default)
        if isinstance(type_, graphql.GraphQLEnumType):
            values = tuple(type_.values)
            return primitives.enum(values, nullable, default=default)
        # Types with children
        if isinstance(type_, graphql.GraphQLList):
            return self.lists(type_, nullable, default=default)
        if isinstance(type_, graphql.GraphQLInputObjectType):
            return self.objects(type_, nullable)
        raise TypeError(f"Type {type_.__class__.__name__} is not supported.")

    @instance_cache(
        lambda type_, nullable=True, default=None: (
            make_type_name(type_),
            nullable,
            default,
        )
    )
    def lists(
        self,
        type_: graphql.GraphQLList,
        nullable: bool = True,
        default: Optional[graphql.ValueNode] = None,
    ) -> st.SearchStrategy[graphql.ListValueNode]:
        """Generate a `graphql.ListValueNode`."""
        strategy = st.lists(self.values(type_.of_type))
        return primitives.list_(strategy, nullable, default=default)

    @instance_cache(lambda type_, nullable=True: (type_.name, nullable))
    def objects(
        self, type_: graphql.GraphQLInputObjectType, nullable: bool = True
    ) -> st.SearchStrategy[graphql.ObjectValueNode]:
        """Generate a `graphql.ObjectValueNode`."""
        fields = {
            name: field
            for name, field in type_.fields.items()
            # Generate optional fields that are possible to generate and all required fields.
            # If a required field is not possible to generate, then it will fail deeper anyway
            if self.can_generate_field(field) or graphql.is_required_input_field(field)
        }
        strategy = subset_of_fields(fields, force_required=True).flatmap(self.lists_of_object_fields)
        return primitives.maybe_null(strategy.map(nodes.Object), nullable)

    def can_generate_field(self, field: graphql.GraphQLInputField) -> bool:
        """Whether it is possible to generate values for the given field."""
        type_ = unwrap_field_type(field)
        return (
            # Can generate any non-scalar
            not isinstance(type_, graphql.GraphQLScalarType)
            # Default scalars
            or type_.name in BUILT_IN_SCALAR_TYPE_NAMES
            # User-provided scalars
            or type_.name in self.custom_scalars
        )

    def lists_of_object_fields(
        self, items: List[Tuple[str, graphql.GraphQLInputField]]
    ) -> st.SearchStrategy[List[graphql.ObjectFieldNode]]:
        return st.tuples(
            *(
                self.values(
                    field.type,
                    field.ast_node.default_value if field.ast_node is not None else None,
                ).map(factories.object_field(name))
                for name, field in items
            )
        ).map(list)

    @instance_cache(
        lambda interface, implementations: (
            interface.name,
            tuple(impl.name for impl in implementations),
        )
    )
    def interfaces(
        self,
        interface: graphql.GraphQLInterfaceType,
        implementations: List[InterfaceOrObject],
    ) -> st.SearchStrategy[SelectionNodes]:
        """Build query for GraphQL interface type."""
        # If there are implementations that have fields with the same name but different types
        # then the resulting query should not have these fields simultaneously
        strategies, overlapping_fields = self.collect_fragment_strategies(implementations)
        if overlapping_fields:
            return compose_interfaces_with_filter(self.selections(interface), strategies, self.schema.type_map)
        # No overlapping - safe to choose any subset of fields within the interface itself and any fragment
        return st.tuples(self.selections(interface), *strategies).map(flatten)  # type: ignore

    @instance_cache(lambda items: tuple(item.name for item in items))
    def inline_fragments(self, items: List[graphql.GraphQLObjectType]) -> st.SearchStrategy[SelectionNodes]:
        """Create inline fragment nodes for each given item."""
        # If there are implementations that have fields with the same name but different types
        # then the resulting query should not have these fields simultaneously
        strategies, overlapping_fields = self.collect_fragment_strategies(items)
        if overlapping_fields:
            return compose_interfaces_with_filter(EMPTY_LISTS_STRATEGY, strategies, self.schema.type_map)
        # No overlapping - safe to choose any subset of fields within the interface itself and any fragment
        return st.tuples(*(self.inline_fragment(type_) for type_ in items)).map(list)

    @instance_cache(lambda type_: type_.name)
    def inline_fragment(self, type_: graphql.GraphQLObjectType) -> st.SearchStrategy[graphql.InlineFragmentNode]:
        """Build `InlineFragmentNode` for the given type."""
        return self.selections(type_).map(factories.inline_fragment(type_.name))

    @instance_cache(lambda type_, fields=None: (type_.name, fields))
    def selections(
        self,
        object_type: InterfaceOrObject,
        fields: Optional[Tuple[str, ...]] = None,
    ) -> st.SearchStrategy[List[graphql.FieldNode]]:
        """Generate a subset of fields defined on the given type."""
        if fields:
            subset = {name: value for name, value in object_type.fields.items() if name in fields}
        else:
            subset = object_type.fields
        # minimum 1 field, an empty query is not valid
        return subset_of_fields(subset).flatmap(self.lists_of_fields)

    def lists_of_fields(self, items: List[Tuple[str, Field]]) -> st.SearchStrategy[List[graphql.FieldNode]]:
        return st.tuples(
            *(
                st.tuples(self.list_of_arguments(field.args), self.selections_for_type(field)).map(
                    factories.field(name)
                )
                for name, field in items
            )
        ).map(list)

    @instance_cache(lambda items: tuple(item.name for item in items))
    def collect_fragment_strategies(
        self, items: List[graphql.GraphQLObjectType]
    ) -> Tuple[List[st.SearchStrategy[graphql.InlineFragmentNode]], bool]:
        field_types: Dict[str, graphql.GraphQLType] = {}
        strategies = []
        has_overlapping_fields = False
        for item in items:
            if not has_overlapping_fields:
                for name, field in item.fields.items():
                    if name in field_types:
                        if not is_equal_type(field.type, field_types[name]):
                            # There are fields with the same names but different types
                            has_overlapping_fields = True
                    else:
                        field_types[name] = field.type
            strategies.append(self.inline_fragment(item))
        return strategies, has_overlapping_fields

    def list_of_arguments(
        self, arguments: Dict[str, graphql.GraphQLArgument]
    ) -> st.SearchStrategy[List[graphql.ArgumentNode]]:
        """Generate a list `graphql.ArgumentNode` for a field."""
        if not arguments:
            return st.just([])

        @st.composite  # type: ignore
        def inner(draw: Any) -> List[graphql.ArgumentNode]:
            args = []
            for name, argument in arguments.items():
                default = argument.ast_node.default_value if argument.ast_node is not None else None
                try:
                    argument_strategy = self.values(argument.type, default=default)
                except InvalidArgument:
                    if not isinstance(argument.type, graphql.GraphQLNonNull):
                        # If the type is nullable, then either generate `null` or skip it completely
                        if draw(st.booleans()):
                            args.append(graphql.ArgumentNode(name=graphql.NameNode(value=name), value=nodes.Null))
                        continue
                    raise
                args.append(draw(argument_strategy.map(factories.argument(name))))
            return args

        return inner()

    def selections_for_type(
        self,
        field: graphql.GraphQLField,
    ) -> st.SearchStrategy[Optional[SelectionNodes]]:
        """Extract proper type from the field and generate field nodes for this type."""
        field_type = unwrap_field_type(field)
        if isinstance(field_type, graphql.GraphQLObjectType):
            return self.selections(field_type)
        if isinstance(field_type, graphql.GraphQLInterfaceType):
            # Besides the fields on the interface type, it is possible to generate inline fragments on types that
            # implement this interface type
            implementations = self.schema.get_implementations(field_type).objects
            if not implementations:
                # Shortcut when there are no implementations - take fields from the interface itself
                return self.selections(field_type)
            return st.lists(st.sampled_from(implementations), min_size=1, unique_by=BY_NAME).flatmap(
                lambda impls: self.interfaces(field_type, impls)
            )
        if isinstance(field_type, graphql.GraphQLUnionType):
            # A union is a set of object types - take a subset of them and generate inline fragments
            return st.lists(st.sampled_from(field_type.types), min_size=1, unique_by=BY_NAME).flatmap(
                self.inline_fragments
            )
        # Other types don't have fields
        return st.none()


def check_nullable(type_: graphql.GraphQLInputType) -> Tuple[graphql.GraphQLInputType, bool]:
    """Get the wrapped type and detect if it is nullable."""
    nullable = True
    if isinstance(type_, graphql.GraphQLNonNull):
        type_ = type_.of_type
        nullable = False
    return type_, nullable


def unwrap_field_type(field: Field) -> graphql.GraphQLNamedType:
    """Get the underlying field type which is not wrapped."""
    type_ = field.type
    while isinstance(type_, graphql.GraphQLWrappingType):
        type_ = type_.of_type
    return type_


def make_type_name(type_: graphql.GraphQLType) -> str:
    """Create a name for a type."""
    name = ""
    while isinstance(type_, graphql.GraphQLWrappingType):
        name += type_.__class__.__name__.replace("GraphQL", "")
        type_ = type_.of_type
    return f"{name}{type_.name}"


@st.composite  # type: ignore
def compose_interfaces_with_filter(
    draw: Any,
    already_selected: st.SearchStrategy[List],
    strategies: List[st.SearchStrategy[SelectionNodes]],
    type_map: Dict[str, graphql.GraphQLType],
) -> SelectionNodes:
    selection_nodes = draw(already_selected)
    # Store what fields are already used and their corresponding types
    seen: Dict[str, graphql.GraphQLType] = {}

    def mark_seen(frag: graphql.InlineFragmentNode) -> None:
        # Add this fragment's fields to `seen`
        fragment_type = type_map[frag.type_condition.name.value]
        for selected in frag.selection_set.selections:
            seen.setdefault(selected.name.value, fragment_type.fields[selected.name.value].type)

    def add_alias(frag: graphql.InlineFragmentNode) -> graphql.InlineFragmentNode:
        # Add an alias for all fields that have the same name with already selected ones but a different type
        fragment_type = type_map[frag.type_condition.name.value]
        for selected in frag.selection_set.selections:
            field_name = selected.name.value
            if field_name in seen:
                field_type = fragment_type.fields[field_name].type
                if not is_equal_type(seen[field_name], field_type):
                    selected.alias = graphql.NameNode(value=f"{field_name}_{make_type_name(field_type)}")
        return frag

    for strategy in strategies:
        fragment = draw(strategy.map(add_alias))
        selection_nodes.append(fragment)
        mark_seen(fragment)
    return selection_nodes


def subset_of_fields(
    fields: Dict[str, graphql.GraphQLInputField], *, force_required: bool = False
) -> st.SearchStrategy[List[Tuple[str, graphql.GraphQLInputField]]]:
    """A helper to select a subset of fields."""
    if not fields:
        # The schema is invalid as there should be at least one field
        # But there should not be an internal error because of it
        return EMPTY_LISTS_STRATEGY
    field_pairs = sorted(fields.items())
    # if we need to always generate required fields, then return them and extend with a subset of optional fields
    if force_required:
        required, optional = [], {}
        for name, field in field_pairs:
            # TYPING: `field` is always `GraphQLInputField` as `force_required` equals `True` only with
            # `GraphQLInputObjectType`. A better solution is to create a separate function
            if graphql.is_required_input_field(field):  # type: ignore
                required.append((name, field))
            else:
                optional[name] = field
        if optional:
            return subset_of_fields(optional).map(required.__add__)
        return st.just(required)
    # pairs are unique by field name
    return st.lists(st.sampled_from(field_pairs), min_size=1, unique_by=lambda x: x[0])


def _make_strategy(
    schema: graphql.GraphQLSchema,
    *,
    type_: graphql.GraphQLObjectType,
    fields: Optional[Iterable[str]] = None,
    custom_scalars: Optional[CustomScalarStrategies] = None,
    alphabet: st.SearchStrategy[str],
    allow_null: bool = True,
) -> st.SearchStrategy[List[graphql.FieldNode]]:
    if fields is not None:
        fields = tuple(fields)
        validation.validate_fields(fields, list(type_.fields))
    if custom_scalars:
        validation.validate_custom_scalars(custom_scalars)
    return GraphQLStrategy(
        schema=schema, alphabet=alphabet, custom_scalars=custom_scalars or {}, allow_null=allow_null
    ).selections(type_, fields=fields)


def _build_alphabet(allow_x00: bool = True, codec: Optional[str] = "utf-8") -> st.SearchStrategy[str]:
    return st.characters(
        codec=codec, min_codepoint=0 if allow_x00 else 1, max_codepoint=0xFFFF, blacklist_categories=["Cs"]
    )


@cacheable  # type: ignore
def queries(
    schema: Union[str, graphql.GraphQLSchema],
    *,
    fields: Optional[Iterable[str]] = None,
    custom_scalars: Optional[CustomScalarStrategies] = None,
    print_ast: AstPrinter = graphql.print_ast,
    allow_x00: bool = True,
    allow_null: bool = True,
    codec: Optional[str] = "utf-8",
) -> st.SearchStrategy[str]:
    r"""A strategy for generating valid queries for the given GraphQL schema.

    The output query will contain a subset of fields defined in the `Query` type.

    :param schema: GraphQL schema as a string or `graphql.GraphQLSchema`.
    :param fields: Restrict generated fields to ones in this list.
    :param custom_scalars: Strategies for generating custom scalars.
    :param print_ast: A function to convert the generated AST to a string.
    :param allow_x00: Determines whether to allow the generation of `\x00` bytes within strings.
    :param allow_null: Whether `null` values should be used for optional arguments.
    :param codec: Specifies the codec used for generating strings.
    """
    parsed_schema = validation.maybe_parse_schema(schema)
    if parsed_schema.query_type is None:
        raise InvalidArgument("Query type is not defined in the schema")
    alphabet = _build_alphabet(allow_x00=allow_x00, codec=codec)
    return (
        _make_strategy(
            parsed_schema,
            type_=parsed_schema.query_type,
            fields=fields,
            custom_scalars=custom_scalars,
            alphabet=alphabet,
            allow_null=allow_null,
        )
        .map(make_query)
        .map(print_ast)
    )


@cacheable  # type: ignore
def mutations(
    schema: Union[str, graphql.GraphQLSchema],
    *,
    fields: Optional[Iterable[str]] = None,
    custom_scalars: Optional[CustomScalarStrategies] = None,
    print_ast: AstPrinter = graphql.print_ast,
    allow_x00: bool = True,
    allow_null: bool = True,
    codec: Optional[str] = "utf-8",
) -> st.SearchStrategy[str]:
    r"""A strategy for generating valid mutations for the given GraphQL schema.

    The output mutation will contain a subset of fields defined in the `Mutation` type.

    :param schema: GraphQL schema as a string or `graphql.GraphQLSchema`.
    :param fields: Restrict generated fields to ones in this list.
    :param custom_scalars: Strategies for generating custom scalars.
    :param print_ast: A function to convert the generated AST to a string.
    :param allow_x00: Determines whether to allow the generation of `\x00` bytes within strings.
    :param allow_null: Whether `null` values should be used for optional arguments.
    :param codec: Specifies the codec used for generating strings.
    """
    parsed_schema = validation.maybe_parse_schema(schema)
    if parsed_schema.mutation_type is None:
        raise InvalidArgument("Mutation type is not defined in the schema")
    alphabet = _build_alphabet(allow_x00=allow_x00, codec=codec)
    return (
        _make_strategy(
            parsed_schema,
            type_=parsed_schema.mutation_type,
            fields=fields,
            custom_scalars=custom_scalars,
            alphabet=alphabet,
            allow_null=allow_null,
        )
        .map(make_mutation)
        .map(print_ast)
    )


@cacheable  # type: ignore
def from_schema(
    schema: Union[str, graphql.GraphQLSchema],
    *,
    fields: Optional[Iterable[str]] = None,
    custom_scalars: Optional[CustomScalarStrategies] = None,
    print_ast: AstPrinter = graphql.print_ast,
    allow_x00: bool = True,
    allow_null: bool = True,
    codec: Optional[str] = "utf-8",
) -> st.SearchStrategy[str]:
    r"""A strategy for generating valid queries and mutations for the given GraphQL schema.

    :param schema: GraphQL schema as a string or `graphql.GraphQLSchema`.
    :param fields: Restrict generated fields to ones in this list.
    :param custom_scalars: Strategies for generating custom scalars.
    :param print_ast: A function to convert the generated AST to a string.
    :param allow_x00: Determines whether to allow the generation of `\x00` bytes within strings.
    :param allow_null: Whether `null` values should be used for optional arguments.
    :param codec: Specifies the codec used for generating strings.
    """
    parsed_schema = validation.maybe_parse_schema(schema)
    if custom_scalars:
        validation.validate_custom_scalars(custom_scalars)
    query = parsed_schema.query_type
    mutation = parsed_schema.mutation_type
    query_fields = None
    mutation_fields = None
    if fields is not None:
        # Split fields based on the type they are defined on & validate them
        fields = tuple(fields)
        available_fields = []
        if query is not None:
            query_fields = tuple(field for field in fields if field in query.fields)
            available_fields.extend(query.fields)
        if mutation is not None:
            mutation_fields = tuple(field for field in fields if field in mutation.fields)
            available_fields.extend(mutation.fields)
        validation.validate_fields(fields, available_fields)

    alphabet = _build_alphabet(allow_x00=allow_x00, codec=codec)
    strategy = GraphQLStrategy(
        parsed_schema, alphabet=alphabet, custom_scalars=custom_scalars or {}, allow_null=allow_null
    )
    strategies = [
        strategy.selections(type_, fields=type_fields).map(node_factory).map(print_ast)
        for (type_, type_fields, node_factory) in (
            (query, query_fields, make_query),
            (mutation, mutation_fields, make_mutation),
        )
        # If a type is defined in the schema and don't have restrictions on fields or has at least one selected field
        if type_ is not None and (type_fields is None or len(type_fields) > 0)
    ]
    if not strategies:
        raise InvalidArgument("Query or Mutation type must be provided")
    return reduce(or_, strategies)



from typing import Dict, List, Optional, Tuple

import graphql

from ..types import SelectionNodes


def maybe_add_alias_to_node(node: graphql.SelectionNode, seen: Dict[Tuple[str, str], List]) -> None:
    if isinstance(node, graphql.FieldNode):
        maybe_add_alias(node, node.arguments, seen)
        if node.selection_set.selections:
            for selection in node.selection_set.selections:
                maybe_add_alias_to_node(selection, seen)
    if isinstance(node, graphql.InlineFragmentNode):
        for selection in node.selection_set.selections:
            maybe_add_alias(selection, selection.arguments, seen)
            if selection.selection_set.selections:
                for sub_selection in selection.selection_set.selections:
                    maybe_add_alias_to_node(sub_selection, seen)


def maybe_add_alias(
    field_node: graphql.FieldNode,
    arguments: List[graphql.ArgumentNode],
    seen: Dict[Tuple[str, str], List],
) -> None:
    for argument in arguments:
        key = (field_node.name.value, argument.name.value)
        value = argument.value
        if key in seen:
            # Simply add an alias, the values could be the same, so it not technically necessary, but this is safe
            # and simpler, but a bit reduces the possible input variety
            field_node.alias = graphql.NameNode(value=f"{field_node.name.value}_{len(seen[key])}")
            seen[key].append(value)
        else:
            seen[key] = [value]


def add_selection_aliases(nodes: Optional[SelectionNodes]) -> Optional[SelectionNodes]:
    """Add aliases to fields that have conflicting argument types."""
    if nodes and len(nodes) > 1:
        seen: Dict[Tuple[str, str], List] = {}
        for node in nodes:
            maybe_add_alias_to_node(node, seen)
    return nodes



import pytest
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.errors import InvalidArgument

from hypothesis_graphql import from_schema

QUERY = """type Query {
  getBooks: [Book]
  getAuthors: [Author]
}"""
MUTATION = """type Mutation {
    addBook(title: String!, author: String!): Book!
    addAuthor(name: String!): Author!
}"""
QUERY_FIELDS = ["getBooks", "getAuthors"]
MUTATION_FIELDS = ["addBook", "addAuthor"]


@pytest.mark.parametrize(
    "types, available_fields",
    (
        (QUERY, QUERY_FIELDS),
        (MUTATION, MUTATION_FIELDS),
        (f"{QUERY}\n{MUTATION}", QUERY_FIELDS + MUTATION_FIELDS),
    ),
)
@given(data=st.data())
def test_from_schema(data, schema, validate_operation, types, available_fields):
    # When the `from_schema` constructor is used
    schema += f"\n{types}"
    selected_fields = data.draw(st.lists(st.sampled_from(available_fields), min_size=1, unique=True))
    other_fields = [field for field in available_fields if field not in selected_fields]
    # Then it should generate both, queries & mutations
    query = data.draw(from_schema(schema, fields=selected_fields))
    for field in other_fields:
        assert field not in query
    # Just Mutation without Query is an invalid schema, but still can generate data
    # Validate other schemas that are assumed to be valid
    if types != MUTATION:
        validate_operation(schema, query)


def test_no_query_no_mutation(schema, validate_operation):
    with pytest.raises(InvalidArgument, match="Query or Mutation type must be provided"):
        from_schema(schema)



import re

import pytest
from hypothesis import find, given
from hypothesis import strategies as st
from hypothesis.errors import InvalidArgument

from hypothesis_graphql import nodes, queries

CUSTOM_SCALAR_TEMPLATE = """
scalar Date

type Object {{
  created: Date
}}

input QueryInput {{
  created: Date
  id: String!
}}

input RequiredQueryInput {{
  created: Date!
}}

type Query {{
  {query}
}}
"""


@given(data=st.data())
def test_custom_scalar_non_argument(data, validate_operation):
    # When a custom scalar type is defined
    # And is used in a non-argument position

    schema = CUSTOM_SCALAR_TEMPLATE.format(query="getObjects: [Object]")
    query = data.draw(queries(schema))
    validate_operation(schema, query)
    # Then queries should be generated
    assert "created" in query


def test_custom_scalar_argument_nullable(validate_operation):
    # When a custom scalar type is defined
    # And is used in an argument position
    # And is nullable
    # And there are no other arguments

    num_of_queries = 0

    schema = CUSTOM_SCALAR_TEMPLATE.format(query="getByDate(created: Date): Object")

    @given(query=queries(schema))
    def test(query):
        nonlocal num_of_queries

        num_of_queries += 1
        validate_operation(schema, query)
        assert "getByDate {" in query or "getByDate(created: null)" in query

    test()
    # Then only two queries should be generated - no fields, and `created: null`
    assert num_of_queries == 2


@pytest.mark.parametrize("input_type", ("Date", "RequiredQueryInput"))
@given(data=st.data())
def test_custom_scalar_argument(data, input_type):
    # When a custom scalar type is defined
    # And is used in an argument position
    # And is not nullable

    schema = CUSTOM_SCALAR_TEMPLATE.format(query=f"getByDate(created: {input_type}!): Object")

    with pytest.raises(TypeError, match="Scalar 'Date' is not supported"):
        data.draw(queries(schema))


@pytest.mark.parametrize("other_type", ("String!", "String"))
def test_custom_scalar_nested_argument(validate_operation, other_type):
    # When a custom scalar type is defined
    # And is used as a field inside an input type
    # And is nullable
    schema = f"""
scalar Date

input QueryInput {{
  created: Date
  id: {other_type}
}}

type Query {{
  getByDate(created: QueryInput!): Int
}}"""

    # Then it could be skipped
    strategy = queries(schema)

    @given(strategy)
    def test(query):
        validate_operation(schema, query)

    test()
    # And "id" is still possible to generate
    assert find(strategy, lambda x: "id" in x).strip() in (
        '{\n  getByDate(created: {id: ""})\n}',
        "{\n  getByDate(created: {id: null})\n}",
    )


@given(data=st.data())
def test_custom_scalar_field(data, validate_operation):
    # When a custom scalar type is defined
    # And is used in a field position
    # And is not nullable
    schema = """
    scalar Date

    type Object {
      created: Date!
    }
    type Query {
      getObject: Object
    }
    """
    # Then query should be generated without errors
    query = data.draw(queries(schema))
    validate_operation(schema, query)
    assert (
        query.strip()
        == """{
  getObject {
    created
  }
}"""
    )


@given(data=st.data())
def test_custom_scalar_registered(data, validate_operation):
    # When the user registered a custom strategy for a scalar
    # Then it should generate valid queries
    schema = CUSTOM_SCALAR_TEMPLATE.format(query="getByDate(created: Date!): Int")
    expected = "EXAMPLE"

    query = data.draw(queries(schema, custom_scalars={"Date": st.just(expected).map(nodes.String)}))
    validate_operation(schema, query)
    assert f'getByDate(created: "{expected}")' in query


@pytest.mark.parametrize(
    "custom_scalars, expected",
    (
        (
            {"Date": 42},
            re.escape(
                r"custom_scalars['Date']=42 must be a Hypothesis strategy which generates AST nodes matching this"
            ),
        ),
        ({42: 42}, "scalar name 42 must be a string"),
    ),
)
def test_invalid_custom_scalar_strategy(custom_scalars, expected):
    # When the user passes `custom_scalars`
    # And it has a wrong type
    # Then there should be an error
    schema = CUSTOM_SCALAR_TEMPLATE.format(query="getByDate(created: Date!): Int")
    with pytest.raises(InvalidArgument, match=expected):
        queries(schema, custom_scalars=custom_scalars)



import typing
from functools import lru_cache

import graphql

# These types are not cache-efficient.
# Constructors are passed as locals to optimize the byte code a little


def String(
    value: typing.Any,
    StringValueNode: typing.Type[graphql.StringValueNode] = graphql.StringValueNode,
) -> graphql.StringValueNode:
    return StringValueNode(value=str(value))


def Float(
    value: float,
    FloatValueNode: typing.Type[graphql.FloatValueNode] = graphql.FloatValueNode,
) -> graphql.FloatValueNode:
    return FloatValueNode(value=str(value))


def Int(value: int, IntValueNode: typing.Type[graphql.IntValueNode] = graphql.IntValueNode) -> graphql.IntValueNode:
    return IntValueNode(value=str(value))


def Object(
    fields: typing.List[graphql.ObjectFieldNode],
    ObjectValueNode: typing.Type[graphql.ObjectValueNode] = graphql.ObjectValueNode,
) -> graphql.ObjectValueNode:
    return ObjectValueNode(fields=fields)


def List(
    values: typing.List[graphql.ValueNode],
    ListValueNode: typing.Type[graphql.ListValueNode] = graphql.ListValueNode,
) -> graphql.ListValueNode:
    return ListValueNode(values=values)


# Boolean & Enum nodes have a limited set of variants, therefore caching is effective in this case


@lru_cache
def Boolean(value: bool) -> graphql.BooleanValueNode:
    return graphql.BooleanValueNode(value=value)


@lru_cache
def Enum(value: str) -> graphql.EnumValueNode:
    return graphql.EnumValueNode(value=value)


Null = graphql.NullValueNode()



import graphql
import pytest
from graphql import GraphQLNamedType
from hypothesis import assume, find, given, settings
from hypothesis import strategies as st
from hypothesis.errors import InvalidArgument

from hypothesis_graphql import nodes, queries
from hypothesis_graphql._strategies.strategy import GraphQLStrategy
from hypothesis_graphql.cache import cached_build_schema


@pytest.fixture(scope="session")
def simple_schema(schema):
    return (
        schema
        + """type Query {
          getBooks: [Book]
          getAuthors: [Author]
        }"""
    )


@pytest.mark.parametrize(
    "query_type",
    (
        """type Query {
      getBooks: [Book]
      getAuthors: [Author]
    }""",
        """type Query {
      getBooksByAuthor(name: String): [Book]
    }""",
    ),
)
@given(data=st.data())
def test_query(data, schema, query_type, validate_operation):
    schema = schema + query_type
    query = data.draw(queries(schema))
    validate_operation(schema, query)


@given(data=st.data())
def test_query_subset(data, simple_schema, validate_operation):
    query = data.draw(queries(simple_schema, fields=["getBooks"]))
    validate_operation(simple_schema, query)
    assert "getAuthors" not in query


def test_empty_fields(simple_schema):
    with pytest.raises(ValueError, match="If you pass `fields`, it should not be empty"):
        queries(simple_schema, fields=[])


def test_wrong_fields(simple_schema):
    with pytest.raises(ValueError, match="Unknown fields: wrong"):
        queries(simple_schema, fields=["wrong"])


@given(data=st.data())
def test_query_from_graphql_schema(data, schema, validate_operation):
    query = """type Query {
      getBooksByAuthor(name: String): [Book]
    }"""
    schema = cached_build_schema(schema + query)
    query = data.draw(queries(schema))
    validate_operation(schema, query)


@pytest.mark.parametrize("allow_null", (True, False))
@pytest.mark.parametrize("notnull", (True, False))
@pytest.mark.parametrize(
    "arguments, node_names",
    (
        ("int: Int", ("IntValueNode",)),
        ("float: Float", ("FloatValueNode",)),
        ("string: String", ("StringValueNode",)),
        ("id: ID", ("IntValueNode", "StringValueNode")),
        ("boolean: Boolean", ("BooleanValueNode",)),
        ("color: Color", ("EnumValueNode",)),
        ("color: EnumInput", ("EnumValueNode",)),
        ("contain: [Int]", ("ListValueNode", "IntValueNode")),
        ("contain: [Int!]", ("ListValueNode", "IntValueNode")),
        ("contain: [Float]", ("ListValueNode", "FloatValueNode")),
        ("contain: [Float!]", ("ListValueNode", "FloatValueNode")),
        ("contain: [String]", ("ListValueNode", "StringValueNode")),
        ("contain: [String!]", ("ListValueNode", "StringValueNode")),
        ("contain: [Boolean]", ("ListValueNode", "BooleanValueNode")),
        ("contain: [Boolean!]", ("ListValueNode", "BooleanValueNode")),
        ("contain: [Color]", ("ListValueNode", "EnumValueNode")),
        ("contain: [Color!]", ("ListValueNode", "EnumValueNode")),
        ("contain: [[Int]]", ("ListValueNode", "IntValueNode")),
        ("contain: [[Int]!]", ("ListValueNode", "IntValueNode")),
        ("contains: QueryInput", ("ObjectValueNode",)),
        ("contains: RequiredInput", ("ObjectValueNode",)),
        ("contains: NestedQueryInput", ("ObjectValueNode",)),
    ),
)
@given(data=st.data())
def test_arguments(data, schema, arguments, node_names, allow_null, notnull, validate_operation):
    if notnull:
        arguments += "!"
    query_type = f"""type Query {{
      getModel({arguments}): Model
    }}"""

    schema = schema + query_type
    query = data.draw(queries(schema, allow_null=allow_null))
    validate_operation(schema, query)
    for node_name in node_names:
        assert node_name not in query
    if notnull:
        assert "getModel(" in query
    if not allow_null:
        assert "null)" not in query
    parsed = graphql.parse(query)
    selection = parsed.definitions[0].selection_set.selections[0]
    if notnull:
        # there should be one argument if it is not null
        assert len(selection.arguments) == 1
    # at least one Model field is selected
    assert len(selection.selection_set.selections) > 0


@pytest.mark.parametrize(
    "query_type",
    (
        "type Query { getModel: Node }",
        "type Query { getModel: Alone }",
    ),
)
@given(data=st.data())
def test_interface(data, schema, query_type, validate_operation):
    schema = schema + query_type
    query = data.draw(queries(schema))
    validate_operation(schema, query)


@pytest.mark.parametrize(
    "query, minimum",
    (
        (
            "getAuthors: [Author]",
            "",
        ),
        (
            "getAuthors(value: Int!): [Author]",
            "(value: 0)",
        ),
        (
            "getAuthors(value: Float!): [Author]",
            "(value: 0.0)",
        ),
        (
            "getAuthors(value: String!): [Author]",
            '(value: "")',
        ),
        (
            "getAuthors(value: Color!): [Author]",
            "(value: RED)",
        ),
    ),
)
def test_minimal_queries(query, schema, minimum):
    schema = schema + f"type Query {{ {query} }}"
    strategy = queries(schema)
    minimal_query = f"""{{
  getAuthors{minimum} {{
    name
  }}
}}"""
    assert find(strategy, lambda x: True).strip() == minimal_query


def test_missing_query():
    schema = """type Author {
      name: String
    }"""
    with pytest.raises(InvalidArgument, match="Query type is not defined in the schema"):
        queries(schema)


def test_unknown_type(simple_schema):
    # If there will be a new input type in `graphql`

    schema = cached_build_schema(simple_schema)

    class NewType(GraphQLNamedType):
        pass

    with pytest.raises(TypeError, match="Type NewType is not supported."):
        GraphQLStrategy(schema, alphabet=st.characters()).values(NewType("Test"))


@given(data=st.data())
def test_no_surrogates(data, validate_operation):
    # Unicode surrogates are not supported by GraphQL spec
    schema = """
    type Query {
        hello(user: String!): String
    }
    """
    query = data.draw(queries(schema))
    document = validate_operation(schema, query)
    argument_node = document.definitions[0].selection_set.selections[0].arguments[0]
    assume(argument_node.name.value == "user")
    value = argument_node.value.value
    value.encode("utf8")


ALIASES_INTERFACE_TWO_TYPES = """interface Conflict {
  id: ID
}

type FloatModel implements Conflict {
  id: ID,
  query: Float!
}

type StringModel implements Conflict {
  id: ID,
  query: String!
}

type Query {
  getData: Conflict
}"""
ALIASES_MULTIPLE_INTERFACE_OVERLAP = """interface Nullable {
  value: Float
}
interface Another {
  non_value: Float
}
interface NotNullable {
  value: Float!
}

type First implements Nullable {
  value: Float
}

type Second implements NotNullable & Another {
  value: Float!
  non_value: Float
}

union FirstOrSecond = First | Second

type Query {
  getData: FirstOrSecond!
}"""
ALIASES_INTERFACE_THREE_TYPES = """interface Conflict {
  id: ID!
}

type First implements Conflict {
  id: ID!
  key: String
}

type Second implements Conflict {
  id: ID!
  key: String
}

type Third implements Conflict {
  id: ID!
  key: [String]
}

type Query {
  getData: Conflict
}"""
ALIASES_INTERFACE_NESTED_TYPE = """interface Conflict {
  keywords: [Keyword!]!
}

type First implements Conflict {
  keywords: [Keyword!]!
}

type Keyword {
  values(first: Int): String!
}

type Query {
  getData(input: Int!): Conflict
}"""
ALIASES_UNION_RETURN_TYPE = """type FloatModel {
  query: Float!
}
type StringModel {
  query: String!
}

union Conflict = FloatModel | StringModel

type Query {
  getData: Conflict
}"""
ALIASES_ARGUMENT_ENUM = """interface Conflict {
  query(arg: Arg): String!
}

type FirstModel implements Conflict {
  query(arg: Arg): String!
}

type SecondModel implements Conflict {
  query(arg: Arg): String!
}

enum Arg {
  First
  Second
}

type Query {
  getConflict(arg: String!): Conflict!
}"""
ALIASES_ARGUMENT_STRING = """interface Conflict {
  query(arg: String): String!
}

type FirstModel implements Conflict {
  query(arg: String): String!
}

type SecondModel implements Conflict {
  query(arg: String): String!
}

type Query {
  getConflict(arg: String!): Conflict!
}"""


@pytest.mark.parametrize(
    "schema",
    (
        ALIASES_INTERFACE_TWO_TYPES,
        ALIASES_INTERFACE_THREE_TYPES,
        ALIASES_INTERFACE_NESTED_TYPE,
        ALIASES_MULTIPLE_INTERFACE_OVERLAP,
        ALIASES_UNION_RETURN_TYPE,
        ALIASES_ARGUMENT_ENUM,
        ALIASES_ARGUMENT_STRING,
    ),
    ids=(
        "interface-two-types",
        "interface-three-types",
        "interface-nested-type",
        "non-interface",
        "union",
        "argument-enum",
        "argument-string",
    ),
)
@given(data=st.data())
def test_aliases(data, validate_operation, schema):
    # See GH-49, GH-57
    # When Query contain types on the same level that have fields with the same name but with different types
    query = data.draw(queries(schema))
    # Then no invalid queries should be generated
    validate_operation(schema, query)


def test_custom_printer(simple_schema):
    def printer(node):
        return str(node)

    @given(queries(simple_schema, print_ast=printer))
    @settings(max_examples=1)
    def test(query):
        assert query == "DocumentNode"

    test()


@pytest.mark.parametrize(
    "type_name, default",
    (
        ("String!", '"foo"'),
        ("[String!]", '["foo"]'),
        ("[String!]!", '["foo"]'),
        ("String", "null"),
        ("ID!", "4432841242"),
        ("ID!", '"Foo"'),
        ("[ID!]", '["Foo"]'),
        ("[ID!]", '["Foo", 42]'),
        ("Int!", "4432841"),
        ("[Int!]", "[4432841]"),
        ("Float!", "4432841242.123"),
        ("[Float!]", "[4432841242.123]"),
        ("Date!", '"2022-04-27"'),
        ("[Date!]", '["2022-04-27"]'),
        # These are kind of useless, but covers some code path
        ("Boolean!", "true"),
        ("Color", "null"),
        ("Color!", "GREEN"),
        ("[Color!]", "[GREEN]"),
        ("[Color!]", "null"),
    ),
)
@pytest.mark.parametrize(
    "format_kwargs",
    (
        lambda x: {
            "inner_type_name": x["type_name"],
            "inner_default": x["default"],
            "outer_type_name": x["type_name"],
            "outer_default": x["default"],
        },
        lambda x: {
            "inner_type_name": x["type_name"],
            "inner_default": x["default"],
            "outer_type_name": "InputData!",
            "outer_default": f'{{ inner: {x["default"]} }}',
        },
    ),
)
def test_default_values(validate_operation, type_name, default, format_kwargs):
    # When the input schema contains nodes with default values
    schema = """
scalar Date

enum Color {{
  RED
  GREEN
  BLUE
}}

input InputData {{
  inner: {inner_type_name} = {inner_default},
}}

type Query {{
  getValue(arg: {outer_type_name} = {outer_default}): Int!
}}
    """.format(**format_kwargs({"type_name": type_name, "default": default}))
    strategy = queries(schema, custom_scalars={"Date": st.just("2022-04-26").map(nodes.String)})
    # Then these default values should be used in generated queries

    all_valid = True

    def validate_and_find(query):
        nonlocal all_valid
        try:
            validate_operation(schema, query)
        except AssertionError:
            all_valid = False
        return default in query

    find(strategy, validate_and_find)
    assert all_valid


@given(data=st.data())
def test_empty_interface(data, validate_operation):
    # When the schema contains an empty interface (which is invalid)
    schema = """
interface Empty

type First implements Empty {
  int: Int!
}
type Second implements Empty {
  float: Float!
}

type Query {
  getByEmpty: Empty
}"""
    # Then query generation should not fail
    query = data.draw(queries(schema))
    # And then schema validation should fail instead
    with pytest.raises(TypeError, match="Type Empty must define one or more fields"):
        validate_operation(schema, query)


@given(data=st.data())
def test_custom_strings(data, validate_operation):
    schema = """
type Query {
  getExample(name: String): String
}"""
    query = data.draw(queries(schema, allow_x00=False, codec="ascii"))
    validate_operation(schema, query)
    assert "\0" not in query
    query.encode("ascii")



"""Set of function for creating GraphQL nodes.

Most of them exist to avoid using lambdas, which might become expensive in Hypothesis in some cases.
"""

from functools import lru_cache
from typing import Callable, List, Optional, Tuple

import graphql

from ..types import SelectionNodes
from .aliases import add_selection_aliases

FieldNodeInput = Tuple[List[graphql.ArgumentNode], Optional[SelectionNodes]]


@lru_cache
def inline_fragment(type_name: str) -> Callable[[SelectionNodes], graphql.InlineFragmentNode]:
    def factory(nodes: SelectionNodes) -> graphql.InlineFragmentNode:
        return graphql.InlineFragmentNode(
            type_condition=graphql.NamedTypeNode(
                name=graphql.NameNode(value=type_name),
            ),
            selection_set=graphql.SelectionSetNode(kind="selection_set", selections=nodes),
        )

    return factory


@lru_cache
def argument(name: str) -> Callable[[graphql.ValueNode], graphql.ArgumentNode]:
    def factory(value: graphql.ValueNode) -> graphql.ArgumentNode:
        return graphql.ArgumentNode(name=graphql.NameNode(value=name), value=value)

    return factory


@lru_cache
def field(name: str) -> Callable[[FieldNodeInput], graphql.FieldNode]:
    def factory(tup: FieldNodeInput) -> graphql.FieldNode:
        return graphql.FieldNode(
            name=graphql.NameNode(value=name),
            arguments=tup[0],
            selection_set=graphql.SelectionSetNode(kind="selection_set", selections=add_selection_aliases(tup[1])),
        )

    return factory


@lru_cache
def object_field(name: str) -> Callable[[graphql.ValueNode], graphql.ObjectFieldNode]:
    def factory(value: graphql.ValueNode) -> graphql.ObjectFieldNode:
        return graphql.ObjectFieldNode(name=graphql.NameNode(value=name), value=value)

    return factory



import graphql
import pytest
from hypothesis import HealthCheck, settings

from hypothesis_graphql.cache import cached_build_schema

settings.register_profile("default", suppress_health_check=[HealthCheck.too_slow], deadline=None)
settings.load_profile("default")


SCHEMA = """
type Book {
  title: String
  author: Author
}

type Author {
  name: String
  books: [Book]
}

enum Color {
  RED
  GREEN
  BLUE
}

input EnumInput {
  color: Color
}

input QueryInput {
  eq: String
  ne: String
}

input RequiredInput {
  eq: Float!,
}

input NestedQueryInput {
  code: QueryInput
}

type Image {
  path: String
}
type Video {
  duration: Int
}

union Media = Image | Video

interface Node {
  id: ID
}

interface Alone {
  id: ID
}

type Model implements Node {
  int: Int,
  float: Float,
  media: Media,
  string: String
  id: ID,
  boolean: Boolean
  color: Color
}
"""


@pytest.fixture(scope="session")
def schema():
    return SCHEMA


@pytest.fixture(scope="session")
def validate_operation():
    def inner(schema, query):
        if isinstance(schema, str):
            parsed_schema = cached_build_schema(schema)
        else:
            parsed_schema = schema
        query_ast = graphql.parse(query)
        errors = graphql.validate(parsed_schema, query_ast)
        for error in errors:
            print(error)
        assert not errors, query
        return query_ast

    return inner
